import com.radzisz.xtrf.model.charge.CATQuantity
import com.radzisz.xtrf.model.project.TaskFinance
import com.radzisz.xtrf.model.CATGrid
import com.radzisz.xtrf.model.dictionary.ActivityType
import com.radzisz.xtrf.model.dictionary.CalculationUnit
import com.radzisz.xtrf.model.dictionary.Service
import com.radzisz.xtrf.model.rate.RateOrigin
import com.radzisz.xtrf.model.project.TaskCATCharge
import com.radzisz.xtrf.model.project.TaskCharge
import com.radzisz.xtrf.model.project.Project
import com.radzisz.xtrf.model.project.ProjectStatus
import com.radzisz.xtrf.model.partner.provider.Provider
import com.radzisz.xtrf.model.partner.provider.ProviderStatus
import com.radzisz.xtrf.model.partner.provider.ProviderPriceProfile
import com.radzisz.xtrf.model.partner.customer.CustomerRate
import com.radzisz.xtrf.model.project.activity.ActivityCATCharge
import com.radzisz.xtrf.model.project.activity.ActivityCharge
import com.radzisz.xtrf.model.charge.CATCharge
import eu.xtrf.integration.cat.domain.WorkflowJobFileType
import com.radzisz.xtrf.model.project.Task
import com.radzisz.xtrf.model.project.TaskStatus
import com.radzisz.xtrf.model.project.activity.Activity
import com.radzisz.xtrf.model.project.activity.ActivityStatus
import com.radzisz.xtrf.model.project.WorkflowJobFile
import com.radzisz.xtrf.utils.TaskCopyOptions
import com.radzisz.xtrf.model.TaskDirectories
import com.radzisz.xtrf.service.AlsoKnownAsService
import com.radzisz.xtrf.service.ProviderService
import groovy.json.JsonOutput
import java.util.Date
import eu.xtrf.cat.MatchType
import com.radzisz.xtrf.model.FixedRateCATGrid
import com.radzisz.xtrf.service.CustomerDataAccessLimitingService
import com.radzisz.xtrf.service.CustomerService
import com.radzisz.xtrf.model.dictionary.Workflow

class BeLazyMacro {
    def list
    def params
    def utils

    BeLazyMacro(list, params, utils) {
        this.list = list
        this.params = params
        this.utils = utils
    }

    def runMacro() {
        def answer = [:]
        try {
            if (params.function == "SetClassicPropertiesAndFinance") {
                answer.jobId = SetClassicPropertiesAndFinance(params)
                answer.success = true
            } else if (params.function == "SetClassicCustomerProjectNumber")  {
                SetClassicCustomerProjectNumber(params)
                answer.success = true
            } else if (params.function == "GetDefaultTasksForClassic") {
                answer.defaultTasksResult = GetDefaultTasksForClassic()
                answer.success = true
            } else if (params.function == "SetSmartFinance") {
                SetSmartFinance(params)
                answer.success = true
            } else if (params.function == "GetVersion") {
                answer.version = 22
                answer.success = true
            } else if (params.function == "SetDefaultRatesOnReceivables") {
                SetDefaultRatesOnReceivables(params)
                answer.success = true
            } else if (params.function == "GetProvidersForJobs") {
                answer.providersForJobIds = GetProvidersForJobs()
                answer.success = true
            } else if (params.function == "GetProviderNamesInExternalSystems") {
                answer.providerNamesInExternalSystems = GetProviderNamesInExternalSystems()
                answer.success = true
            } else if (params.function == "GetOutputFilesForClassic") {
                answer.taskOutputFiles = GetOutputFilesForClassic()
                answer.success = true
            } else if (params.function == "GetPriceProfilesForCustomers") {
                answer.priceProfilesForCustomers = GetPriceProfilesForCustomers(params)
                answer.success = true
            } else if (params.function == "SetPriceProfileForProject") {
                SetPriceProfileForProject(params)
                answer.success = true
            } else if (params.function == "GetJobStatuses") {
                answer.jobStatusesForTasks = GetJobStatuses()
                answer.success = true
            } else if (params.function == "SetVendorRatesOnPayables") {
                SetVendorRatesOnPayables(params)
                answer.success = true
            } else if (params.function == "GetServicesWithWorkflowSteps") {
                answer.services = GetServicesWithWorkflowSteps()
                answer.success = true
            } else if (params.function == "CloseClassicProject") {
                answer.projectStatus = CloseClassicProject()
                answer.success = true
            } else if (params.function == "GetVendors") {
                answer.vendors = GetVendors(params)
                answer.success = true
            } else {
                answer.error = "This function doesn't exist."
                answer.success = false
            }
        } catch (Throwable e) {
            answer.error = e.getMessage()
            answer.success = false
        }
        return JsonOutput.toJson(answer)
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Classic projects related functions
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    ////////////////////////////////////////////////////////////
    // Set the customer project number
    ////////////////////////////////////////////////////////////

    def SetClassicCustomerProjectNumber(def params) {
        list.each { project ->
            if (params.customerProjectNumber != null)
                project.setCustomerProjectNumber(params.customerProjectNumber)
        }
    }

    ////////////////////////////////////////////////////////////
    // Set finance and properties on classic project
    ////////////////////////////////////////////////////////////

    def SetClassicPropertiesAndFinance(def params) {
        def activityId
        list.each { project ->
            def task = project.getTasks().find { it.getId() == params.taskId }
            def taskFinance = task.getProjectFinance()

            def activity = task.getActivities().find { it.getWorkflowJobInstance().getJobPosition() == params.jobPosition }
            def defaultDeadline = activity.getDeadline()
            if (params.createNewJob) {
                def actService = utils.getService("com.radzisz.xtrf.service.ActivityService")
                def workflowJobInstance  = activity.getWorkflowJobInstance()
                activity = actService.copyActivity(activity, task, workflowJobInstance, null, true, false, false, false)
            }
            
            setActivityProperties(activity, params.jobParams, taskFinance, defaultDeadline)
            activityId = activity.getId()
        }
        return activityId
    }

    def setActivityProperties(def activity, def jobParams, def taskFinance, def defaultDeadline) {
        def simpleFinance = (jobParams.analysis == null) || (jobParams.analysis.size() == 0)
        def baseReceivableRate
        def basePayableRate
        def receivableRateOrigin
        def payableRateOrigin

        def calcService = utils.getService("com.radzisz.xtrf.service.dictionary.CalculationUnitService")
        def calculationUnit = calcService.getById(jobParams.calculationUnitId)

        if (jobParams.receivableRate == null) {
            baseReceivableRate = getDefaultRate(activity.getTask(), activity.getActivityType(), calculationUnit)
            receivableRateOrigin = RateOrigin.PRICE_PROFILE
        } else {
            baseReceivableRate = BigDecimal.valueOf(jobParams.receivableRate)
            receivableRateOrigin = RateOrigin.FILLED_MANUALLY
        }

        if (jobParams.payableRate == null) {
            basePayableRate = getDefaultRate(activity.getTask(), activity.getActivityType(), calculationUnit)
            payableRateOrigin = RateOrigin.PRICE_PROFILE
        } else {
            basePayableRate = BigDecimal.valueOf(jobParams.payableRate)
            payableRateOrigin = RateOrigin.FILLED_MANUALLY
        }

        setActivityDeadline(activity, jobParams, defaultDeadline)

        setJobInstructions(activity, jobParams.instructions)

        if (simpleFinance)
            createSimpleFinanceForActivity(baseReceivableRate, basePayableRate, receivableRateOrigin, payableRateOrigin, calculationUnit, activity, taskFinance, jobParams)
        else
            createCatFinanceForActivity(baseReceivableRate, basePayableRate, receivableRateOrigin, payableRateOrigin, calculationUnit, activity, taskFinance, jobParams)
    }

    def setActivityDeadline(def activity, def jobParams, def defaultDeadline)
    {
        if (jobParams.deadline != null && jobParams.deadline != 0) {
            def deadline = new Date(jobParams.deadline)
            if(jobParams.isFirstItemForJob || defaultDeadline == null || defaultDeadline > deadline)
                activity.setDeadline(deadline)
            else
                activity.setDeadline(defaultDeadline)
        } else
        {
            activity.setDeadline(defaultDeadline)
        }
    }

    def setJobInstructions(def activity, def instructions)
    {
        if (instructions == null)
            return

        if (instructions.fromCustomer != null)
        {
            def originInstruction = activity.getCustomerSpecialInstructions()
            activity.setCustomerSpecialInstructions(concatInstructions(originInstruction, instructions.fromCustomer))
        }

        if (instructions.forProvider != null)
        {
            def originInstruction = activity.getProviderSpecialInstructions()
            activity.setProviderSpecialInstructions(concatInstructions(originInstruction, instructions.forProvider))
        }

        if (instructions.internal != null)
        {
            def originInstruction = activity.getInternalSpecialInstructions()
            activity.setInternalSpecialInstructions(concatInstructions(originInstruction, instructions.internal))
        }

        if (instructions.paymentNoteForCustomer != null)
        {
            def originInstruction = activity.getPaymentNote()
            activity.setPaymentNote(concatInstructions(originInstruction, instructions.paymentNoteForCustomer))
        }
    }

    def concatInstructions(def originInstruction, def newInstruction)
    {
        if(originInstruction?.trim())
        {
            return """$originInstruction
                    |$newInstruction""".stripMargin()
        }
        else
        {
            return newInstruction
        }
    }

    def createCatFinanceForActivity(def baseReceivableRate, def basePayableRate, def receivableRateOrigin, def payableRateOrigin, def calculationUnit, def activity, def taskFinance, def jobParams)
    {
            if (jobParams.isFixedRateAnalysis)
                return createFixedRateCatFinanceForActivity(baseReceivableRate, basePayableRate, receivableRateOrigin, payableRateOrigin, calculationUnit, activity, taskFinance, jobParams)
            else
                return createPercentageCatFinanceForActivity(baseReceivableRate, basePayableRate, receivableRateOrigin, payableRateOrigin, calculationUnit, activity, taskFinance, jobParams)
    }

    def createPercentageCatFinanceForActivity(def baseReceivableRate, def basePayableRate, def receivableRateOrigin, def payableRateOrigin, def calculationUnit, def activity, def taskFinance, def jobParams)
    {
        def catQuantity = createCatQuantity(jobParams)

        def defaultCatGrid = createDefaultCatGrid(activity.getTask().getCustomerPriceProfile())
        def payableCatGrid = defaultCatGrid
        def receivableCatGrid = (!jobParams.useCustomWeightsAlways && jobParams.receivableRate == null)
            ? defaultCatGrid
            : createCatGrid(jobParams);

        if (jobParams.createPayable == null || jobParams.createPayable)
        {
            def catPayable = new ActivityCATCharge(calculationUnit, basePayableRate, payableRateOrigin, jobParams.payableDescription, activity)
            catPayable.setCATChargeCATGrid(payableCatGrid)
            catPayable.setCATChargeCATQuantity(catQuantity)
            activity.addCATPayable(catPayable)
        }

        if(jobParams.createReceivable == null || jobParams.createReceivable)
        {
            def receivableJobType
            if (jobParams.receivableJobType != null)
                receivableJobType = getJobType(jobParams.receivableJobType)
            else
                receivableJobType = activity.getActivityType()
            def rate = new CustomerRate(baseReceivableRate, calculationUnit, receivableJobType)
            def catReceivable = new TaskCATCharge(receivableJobType, calculationUnit, jobParams.receivableDescription, rate, receivableCatGrid, catQuantity, taskFinance)
            taskFinance.addCATReceivable(catReceivable)
        }
    }

    def createFixedRateCatFinanceForActivity(def baseReceivableRate, def basePayableRate, def receivableRateOrigin, def payableRateOrigin, def calculationUnit, def activity, def taskFinance, def jobParams)
    {
        def catQuantity = createCatQuantity(jobParams)

        def payableRate = getPayableBasicRateFromAnalysis(jobParams)
        if(payableRate == null)
            payableRate = basePayableRate
        def fixedRateAnalysisForPayable = createFixedRateCATGridForPayable(jobParams)
        if (jobParams.createPayable == null || jobParams.createPayable)
        {
            def catPayable = new ActivityCATCharge(calculationUnit, payableRate, payableRateOrigin, jobParams.payableDescription, activity)
            catPayable.setCATChargeCATQuantity(catQuantity)
            catPayable.setCATChargeFixedRateCATGrid(fixedRateAnalysisForPayable)
            activity.addCATPayable(catPayable)
        }

        if(jobParams.createReceivable == null || jobParams.createReceivable)
        {
            def receivableRate = getReceivableBasicRateFromAnalysis(jobParams)
            if(receivableRate == null)
                receivableRate = baseReceivableRate
            def receivableJobType
            if (jobParams.receivableJobType != null)
                receivableJobType = getJobType(jobParams.receivableJobType)
            else
                receivableJobType = activity.getActivityType()
            def rate = new CustomerRate(receivableRate, calculationUnit, receivableJobType)
            def fixedRateAnalysisForReceivable = createFixedRateCATGridForReceivable(jobParams)
            def catReceivable = new TaskCATCharge(receivableJobType, calculationUnit, jobParams.receivableDescription, rate, new CATGrid(), catQuantity, taskFinance)
            catReceivable.setCATChargeFixedRateCATGrid(fixedRateAnalysisForReceivable)
            taskFinance.addCATReceivable(catReceivable)
        }
    }

    def createSimpleFinanceForActivity(def baseReceivableRate, def basePayableRate, def receivableRateOrigin, def payableRateOrigin, def calculationUnit, def activity, def taskFinance, def jobParams)
    {
        if(jobParams.createPayable == null || jobParams.createPayable)
        {
            def simplePayable = new ActivityCharge(basePayableRate, payableRateOrigin, BigDecimal.valueOf(jobParams.weightedQuantity), calculationUnit, jobParams.payableDescription, activity)
            activity.addPayable(simplePayable)
        }

        if(jobParams.createReceivable == null || jobParams.createReceivable)
        {
            def receivableJobType
            if (jobParams.receivableJobType != null)
                receivableJobType = getJobType(jobParams.receivableJobType)
            else
                receivableJobType = activity.getActivityType()
            def simpleReceivable = new TaskCharge(receivableJobType, calculationUnit, jobParams.receivableDescription, BigDecimal.valueOf(jobParams.weightedQuantity), baseReceivableRate, receivableRateOrigin, taskFinance)
            taskFinance.addReceivable(simpleReceivable)
        }
    }

    def getJobType(def jobTypeId)
    {
        def jobTypeService = utils.getService("com.radzisz.xtrf.service.dictionary.ActivityTypeService")
        return jobTypeService.getById(jobTypeId)
    }

    def createDefaultCatGrid(def priceProfile) {
        def tmRatesList = priceProfile.getTmRates()
        def tmRates = tmRatesList.find{ it.key.toString() == "TRADOS" }?.value

        if (tmRates == null)
            throw new Exception("OtherCatToolNotFound")

        def rates = tmRates.getRates()
        def catGrid = new CATGrid()

        catGrid.setXtranslated(rates[MatchType.X_TRANSLATED].asBigDecimal())
        catGrid.setRepetitions(rates[MatchType.REPETITIONS].asBigDecimal())
        catGrid.setPercent100(rates[MatchType.PERCENT_100].asBigDecimal())
        catGrid.setPercent95_99(rates[MatchType.PERCENT_95_99].asBigDecimal())
        catGrid.setPercent85_94(rates[MatchType.PERCENT_85_94].asBigDecimal())
        catGrid.setPercent75_84(rates[MatchType.PERCENT_75_84].asBigDecimal())
        catGrid.setPercent50_74(rates[MatchType.PERCENT_50_74].asBigDecimal())
        catGrid.setNoMatch(rates[MatchType.NO_MATCH].asBigDecimal())

        return catGrid
    }

    def markProjectAsIncomplete(def project) {
        def currentName = project.getName()
        if (!currentName.endsWith(" - INCOMPLETE!"))
            project.setName(currentName + " - INCOMPLETE!")
    }

    ////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////

    ////////////////////////////////////////////////////////////
    // Get the default tasks created in a classic project
    ////////////////////////////////////////////////////////////

    def GetDefaultTasksForClassic() {
        def result = [:]
        list.each { project ->
            def tasks = project.getTasks()
            tasks = tasks.findAll { it.getLanguageCombination() != null && it.getLanguageCombination().getTargetLanguage() != null }
            if (tasks.isEmpty()) {
                markProjectAsIncomplete(project)
                return null
            }

            result.workflowId = tasks[0].getWorkflow().getId()
            tasks.each { task ->
                def activities = task.getActivities()
                activities.each { activity ->
                    while(activity.isPreviousStatusPossible())
                        activity.setPreviousStatus()
                }
            }
            result.defaultTasks = tasks.collect { [ id: it.getId(), targetLangId: it.getLanguageCombination().getTargetLanguage().getId() ] }
        }
        return result
    }

    ////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////

    ////////////////////////////////////////////////////////////
    // Get providers for jobs in a classic project
    ////////////////////////////////////////////////////////////

    def GetProvidersForJobs() {
        def providersForJobIds = [:]
        list.each { project ->
            def activities = project.getActivities()
            activities.each { activity ->
                def provider = activity.getProvider()
                if(provider != null)
                    providersForJobIds[activity.getId().toString()] = getVendorData(provider)
            }
        }
        return providersForJobIds
    }

    ////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////

    ////////////////////////////////////////////////////////////
    // Get output files for jobs in a classic project
    ////////////////////////////////////////////////////////////

    def GetOutputFilesForClassic() {
        def taskOutputFiles = [:]
        list.each { project ->
            def tasks = project.getTasks()
            tasks.each { task ->
                def taskOutputFilesDto = createTaskOutputFilesDto(task)
                taskOutputFiles[task.getId().toString()] = taskOutputFilesDto
            }
        }

        return taskOutputFiles
    }

    def createTaskOutputFilesDto(def task)
    {
        def outputFilesForJobs = [:]
        def activities = task.getActivities()
        def lastActivityId = null
        activities.each { activity ->
            def outputFilesDto = []
            def outputFiles = activity.getOutFiles()
            outputFiles.each { file ->
                outputFilesDto.push(createFileDto(file))
            }

            lastActivityId = activity.getId().toString()
            outputFilesForJobs[lastActivityId] = outputFilesDto
        }

        def finalOutputFiles = task.getOutputFiles()
        finalOutputFiles.each { file ->
            outputFilesForJobs[lastActivityId].push(createFileDto(file))
        }

        return [
            outputFilesForJobIds: outputFilesForJobs
        ]
    }

    def createFileDto(def file)
    {
        return [
            externalId: file.getId(),
            name: file.getName(),
        ]
    }


    ////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////

    ////////////////////////////////////////////////////////////
    // Get job statuses for Project
    ////////////////////////////////////////////////////////////

    def GetJobStatuses() {
        def jobStatusesForTasks = [:]
        list.each { project ->
        def tasks = project.getTasks()
        tasks.each { task ->
                def taskJobStatuses = [:]
                taskJobStatuses.jobStatuses = getJobStatusesForTask(task)
                jobStatusesForTasks[task.getId().toString()] = taskJobStatuses
            }
        }

        return jobStatusesForTasks
    }



    def getJobStatusesForTask(def task) {
        def jobStatuses = [:]
        def activities = task.getActivities()
        activities.each { activity ->
            jobStatuses[activity.getId().toString()] = activity.getStatus()
        }

        return jobStatuses
    }


    ////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////


    ////////////////////////////////////////////////////////////
    // Close classic project
    ////////////////////////////////////////////////////////////

    def CloseClassicProject() {
        list.each{Project project -> closeProject(project)}
        return list[0].getStatus()
    }

    def closeProject(Project project){
        ProjectStatus status = project.getStatus()
        if(status == ProjectStatus.CLOSED){
            return;
        }

        List<Task> tasks = project.getTasks()
        tasks.each{Task task -> closeTask(task)}
    }
        
    def closeTask(Task task) {
        TaskStatus status = task.getStatus()
        if(status == TaskStatus.CLOSED){
            return;
        }

        List<Activity> activities = []
        activities.addAll(task.getActivities())
        activities.each{
            Activity activity -> closeJob(activity)
        }
        task.close(true, true)
    }
        
    def closeJob(Activity activity){
        ActivityStatus status = activity.getStatus()
        def id = activity.getId()
        if(status == ActivityStatus.ACCEPTED){
            activity.setStatus(ActivityStatus.STARTED)
            activity.setStatus(ActivityStatus.READY)
        }
        if(status == ActivityStatus.STARTED){
            activity.setStatus(ActivityStatus.READY)
        }
    }

    ////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////


    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////



    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Smart projects related functions
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    ////////////////////////////////////////////////////////////
    // Set finance details on receivables
    ////////////////////////////////////////////////////////////

    def SetSmartFinance(def params) {
        list.each { project ->
            if (params.isProjectIncomplete)
                project.setName(project.getName() + " - INCOMPLETE!")
            def tasks = project.getTasks()
            tasks.each { task ->
                def tf = task.getProjectFinance()
                def catReceivables = tf.getCatReceivables()
                catReceivables.findAll { it.isCAT() }.each { receivable ->   
                    def id = Long.toString(receivable.getId())
                    def paramsForReceivable = params.analysesForReceivables[id]
                    if (paramsForReceivable != null) {
                        if(paramsForReceivable.isFixedRateAnalysis)
                        {
                            def fixedRates = createFixedRateCATGridForReceivable(paramsForReceivable)
                            def noMatchFixedRateReceivable = getReceivableBasicRateFromAnalysis(paramsForReceivable)
                            if(noMatchFixedRateReceivable != null)
                                receivable.setRate(noMatchFixedRateReceivable, RateOrigin.FILLED_MANUALLY)
                            receivable.setCATChargeFixedRateCATGrid(fixedRates)
                        }
                        else
                        {
                            def catGrid = createCatGrid(paramsForReceivable)
                            receivable.setCATChargeCATGrid(catGrid)
                        }
                    }
                }
            }
        }
    }

    ////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////

    ////////////////////////////////////////////////////////////
    // Set default rates on receivables
    ////////////////////////////////////////////////////////////

    def SetDefaultRatesOnReceivables(def params) {
        def receivableIds = params.receivableIds.collect { (long)it }
        list.each { project ->
            def tasks = project.getTasks()
            tasks.each { task ->
                def tf = task.getProjectFinance()
                def receivables = tf.getAllReceivables()
                receivables.findAll { it.getId() in receivableIds }.each { receivable ->
                    def defaultRate = getDefaultRate(task, receivable.getActivityType(), receivable.getCalculationUnit())
                    receivable.setRate(defaultRate, RateOrigin.PRICE_PROFILE)
                }
            }
        }
    }

    def SetVendorRatesOnPayables(def params) {
        def payableIds = params.payableIds.collect { (long)it }
        list.each { project ->
            def activities = project.getActivities()
            activities.each { activity ->
                def providerPriceProfile = activity.getProviderPriceProfile()
                if (providerPriceProfile != null)
                {
                    def languageCombination = activity.getLanguageCombination()
                    if (languageCombination != null)
                    {
                        def catPayables = activity.getCatPayables()
                        catPayables.findAll { it.getId() in payableIds }.each { catPayable ->
                            if (catPayable.isPercentage())
                            {
                                def defaultCatGrid = createDefaultCatGrid(providerPriceProfile)
                                catPayable.setInternalCATChargeGrid(defaultCatGrid)
                            }
                        }

                        def payables = activity.getAllPayables()
                        payables.findAll { it.getId() in payableIds }.each { payable ->
                            def rate = providerPriceProfile.getRate(languageCombination.getSourceLanguage(), languageCombination.getTargetLanguage(), activity.getSpecialization(), payable.getCalculationUnit(), activity.getActivityType())
                            if (rate != null && rate.isPresent())
                            {
                                payable.setRate(rate.get())
                            }
                        }
                    }
                }
            }
        }
    }

    ////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////



    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Shared functions
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    def GetServicesWithWorkflowSteps()
    {
        def serviceService = utils.getService("com.radzisz.xtrf.service.dictionary.ServiceService")
        def services = serviceService.getAllActive()
        def servicesDtos = []
        services.each { service ->
            def servicesDto = [
                id: service.getId(),
                name: service.getName(),
                isClassic: service.isClassicService(),
                workflow: GetWorkflowDto(service)
            ]
            servicesDtos.push(servicesDto)
        }
        return servicesDtos
    }

    def GetWorkflowDto(def service)
    {
        if(service.isClassicService()) {
            def workflow = service.getWorkflow()
            if(workflow.isProjectWorkflow())
                workflow = workflow.getDefaultTaskWorkflow()

            if (workflow == null)
                return null
            def workflowSteps = workflow.getAllWorkflowJobs()
            def workflowStepDtos = []
            workflowSteps.each { step ->
                def workflowStepDto = [
                    id: step.getId(),
                    name: step.getName(),
                    activityType: step.getActivityType()?.getName()
                ]
                workflowStepDtos.push(workflowStepDto)
            }

            def workflowDto = [
                id: workflow.getId(),
                name: workflow.getName(),
                workflowSteps: workflowStepDtos
            ]
            return workflowDto
        }
        else {
            def processTemplate = service.getProcessTemplate()
            def workflowDto = [
                id: processTemplate.getProcessId(),
                name: processTemplate.getName(),
                workflowSteps: null
            ]
            return workflowDto
        }
    }

    def GetProviderNamesInExternalSystems()
    {
        def alsoKnownAsService = utils.getService("com.radzisz.xtrf.service.AlsoKnownAsService")
        def alsoKnownAsItems = alsoKnownAsService.getAll()
        def alsoKnownAsDictionary = []
        alsoKnownAsItems.each { alsoKnownAs ->
            if(alsoKnownAs.getProvider() != null)
            {
                def providerNameInExternalSystem = [
                    externalSystemName: alsoKnownAs.getName(),
                    vendorId: alsoKnownAs.getProvider().getId(),
                    nameInExternalSystem: alsoKnownAs.getValue()
                ]
                alsoKnownAsDictionary.push(providerNameInExternalSystem)
            }
        }
        return alsoKnownAsDictionary
    }

    def GetPriceProfilesForCustomers(def params)
    {
        def customerService = utils.getService("com.radzisz.xtrf.service.CustomerService")
        def customers = customerService.getAll()
        
        def customerLimit = params?.customerLimit ?: 5000
        def priceProfilesForCustomers = [:]
        def customersWithBeLazyCategories = [:]

        customers.each { customer ->

            def categories = customer.getCategoriesAsArray()
            categories.each { category ->
                if(category.getDisplayName().compareToIgnoreCase("BeLazy") == 0 || 
                    category.getDisplayName().compareToIgnoreCase("Automation") == 0)
                {
                    customersWithBeLazyCategories[customer.getId()] = true
                }
            }

            priceProfilesForCustomers[customer.getId()] = []
            def priceProfiles = customer.getPriceProfiles()
            priceProfiles.each { priceProfile ->
                def priceProfileValue = [:]
                def currency = [:]
                currency.symbol = priceProfile.getCurrency().getSymbol()
                currency.isoCode = priceProfile.getCurrency().getISOCode()
                currency.name = priceProfile.getCurrency().getName()
                priceProfileValue.id = priceProfile.getId()
                priceProfileValue.name = priceProfile.getName()
                priceProfileValue.currency = currency
                priceProfilesForCustomers[customer.getId()].add(priceProfileValue)
            }
        }

        def filterByCategories = customersWithBeLazyCategories.size() > 0
        if(priceProfilesForCustomers.size() > customerLimit && filterByCategories)
        {
            priceProfilesForCustomers = priceProfilesForCustomers.findAll{ customersWithBeLazyCategories[it.key] == true  }
        }

        return priceProfilesForCustomers.take(customerLimit+1)
    }

    def SetPriceProfileForProject(def params)
    {
        list.each { project ->
            def customer = project. getCustomer()
            def priceProfile = customer.getPriceProfiles().find{ it.getId() == params.priceProfileId }

            if(priceProfile == null)
                throw new Exception("PriceProfileIsNotFoundForCustomer")
            project.setCustomerPriceProfile(priceProfile)
        }
    }


    def createCatGrid(def params) {
        def catGrid = new CATGrid()
        if (params.analysis['X translated'] != null)
            catGrid.setXtranslated(BigDecimal.valueOf(params.analysis['X translated'].percentOfRate))
        if (params.analysis['Repetitions'] != null)
            catGrid.setRepetitions(BigDecimal.valueOf(params.analysis['Repetitions'].percentOfRate))
        if (params.analysis['100%'] != null)
            catGrid.setPercent100(BigDecimal.valueOf(params.analysis['100%'].percentOfRate))
        if (params.analysis['95-99%'] != null)
            catGrid.setPercent95_99(BigDecimal.valueOf(params.analysis['95-99%'].percentOfRate))
        if (params.analysis['85-94%'] != null)
            catGrid.setPercent85_94(BigDecimal.valueOf(params.analysis['85-94%'].percentOfRate))
        if (params.analysis['75-84%'] != null)
            catGrid.setPercent75_84(BigDecimal.valueOf(params.analysis['75-84%'].percentOfRate))
        if (params.analysis['50-74%'] != null)
            catGrid.setPercent50_74(BigDecimal.valueOf(params.analysis['50-74%'].percentOfRate))
        if (params.analysis['No match'] != null)
            catGrid.setNoMatch(BigDecimal.valueOf(params.analysis['No match'].percentOfRate))
        return catGrid
    }

    def createCatQuantity(def params) {
        def catQuantity = new CATQuantity()
        if (params.analysis['X translated'] != null)
            catQuantity.setXtranslated(BigDecimal.valueOf(params.analysis['X translated'].quantity))
        if (params.analysis['Repetitions'] != null)
            catQuantity.setRepetitions(BigDecimal.valueOf(params.analysis['Repetitions'].quantity))
        if (params.analysis['100%'] != null)
            catQuantity.setPercent100(BigDecimal.valueOf(params.analysis['100%'].quantity))
        if (params.analysis['95-99%'] != null)
            catQuantity.setPercent95_99(BigDecimal.valueOf(params.analysis['95-99%'].quantity))
        if (params.analysis['85-94%'] != null)
            catQuantity.setPercent85_94(BigDecimal.valueOf(params.analysis['85-94%'].quantity))
        if (params.analysis['75-84%'] != null)
            catQuantity.setPercent75_84(BigDecimal.valueOf(params.analysis['75-84%'].quantity))
        if (params.analysis['50-74%'] != null)
            catQuantity.setPercent50_74(BigDecimal.valueOf(params.analysis['50-74%'].quantity))
        if (params.analysis['No match'] != null)
            catQuantity.setNoMatch(BigDecimal.valueOf(params.analysis['No match'].quantity))
        return catQuantity
    }

    def createFixedRateCATGridForReceivable(def params) {
        def fixedRateCATGrid = new FixedRateCATGrid()
        if (params.analysis['X translated'] != null)
            fixedRateCATGrid.setXtranslatedRate(BigDecimal.valueOf(params.analysis['X translated'].receivableRate))
        if (params.analysis['Repetitions'] != null)
            fixedRateCATGrid.setRepetitionsRate(BigDecimal.valueOf(params.analysis['Repetitions'].receivableRate))
        if (params.analysis['100%'] != null)
            fixedRateCATGrid.setPercent100Rate(BigDecimal.valueOf(params.analysis['100%'].receivableRate))
        if (params.analysis['95-99%'] != null)
            fixedRateCATGrid.setPercent95_99Rate(BigDecimal.valueOf(params.analysis['95-99%'].receivableRate))
        if (params.analysis['85-94%'] != null)
            fixedRateCATGrid.setPercent85_94Rate(BigDecimal.valueOf(params.analysis['85-94%'].receivableRate))
        if (params.analysis['75-84%'] != null)
            fixedRateCATGrid.setPercent75_84Rate(BigDecimal.valueOf(params.analysis['75-84%'].receivableRate))
        if (params.analysis['50-74%'] != null)
            fixedRateCATGrid.setPercent50_74Rate(BigDecimal.valueOf(params.analysis['50-74%'].receivableRate))
        return fixedRateCATGrid
    }

    def createFixedRateCATGridForPayable(def params) {
        def fixedRateCATGrid = new FixedRateCATGrid()
        if (params.analysis['X translated'] != null)
            fixedRateCATGrid.setXtranslatedRate(BigDecimal.valueOf(params.analysis['X translated'].payableRate))
        if (params.analysis['Repetitions'] != null)
            fixedRateCATGrid.setRepetitionsRate(BigDecimal.valueOf(params.analysis['Repetitions'].payableRate))
        if (params.analysis['100%'] != null)
            fixedRateCATGrid.setPercent100Rate(BigDecimal.valueOf(params.analysis['100%'].payableRate))
        if (params.analysis['95-99%'] != null)
            fixedRateCATGrid.setPercent95_99Rate(BigDecimal.valueOf(params.analysis['95-99%'].payableRate))
        if (params.analysis['85-94%'] != null)
            fixedRateCATGrid.setPercent85_94Rate(BigDecimal.valueOf(params.analysis['85-94%'].payableRate))
        if (params.analysis['75-84%'] != null)
            fixedRateCATGrid.setPercent75_84Rate(BigDecimal.valueOf(params.analysis['75-84%'].payableRate))
        if (params.analysis['50-74%'] != null)
            fixedRateCATGrid.setPercent50_74Rate(BigDecimal.valueOf(params.analysis['50-74%'].payableRate))
        return fixedRateCATGrid
    }

    def getDefaultRate(def task, def activityType, def calculationUnit) {
        try {
            def priceProfile = task.getCustomerPriceProfile()
            if (priceProfile == null)
                return 0
            def rate = priceProfile.getRate(task.getSourceLanguage(), task.getTargetLanguage(), task.getLanguageSpecialization(), calculationUnit, activityType)
            if (rate.isPresent()) {
                def custRate = rate.get()
                return custRate.getRate()
            } else {
                return 0
            }
        } catch (Exception e) {
            return 0
        }
    }

    def getReceivableBasicRateFromAnalysis(def jobParams)
    {
        if (jobParams.analysis['No match'] != null)
            return BigDecimal.valueOf(jobParams.analysis['No match'].receivableRate)
        if (jobParams.analysis['50-74%'] != null)
            return BigDecimal.valueOf(jobParams.analysis['50-74%'].receivableRate)
        if (jobParams.analysis['75-84%'] != null)
            return BigDecimal.valueOf(jobParams.analysis['75-84%'].receivableRate)
        if (jobParams.analysis['85-94%'] != null)
            return BigDecimal.valueOf(jobParams.analysis['85-94%'].receivableRate)
        if (jobParams.analysis['95-99%'] != null)
            return BigDecimal.valueOf(jobParams.analysis['95-99%'].receivableRate)
        if (jobParams.analysis['100%'] != null)
            return BigDecimal.valueOf(jobParams.analysis['100%'].receivableRate)
        if (jobParams.analysis['Repetitions'] != null)
            return BigDecimal.valueOf(jobParams.analysis['Repetitions'].receivableRate)
        if (jobParams.analysis['X translated'] != null)
            return BigDecimal.valueOf(jobParams.analysis['X translated'].receivableRate)
    }

    def getPayableBasicRateFromAnalysis(def jobParams)
    {
        if (jobParams.analysis['No match'] != null)
            return BigDecimal.valueOf(jobParams.analysis['No match'].payableRate)
        if (jobParams.analysis['50-74%'] != null)
            return BigDecimal.valueOf(jobParams.analysis['50-74%'].payableRate)
        if (jobParams.analysis['75-84%'] != null)
            return BigDecimal.valueOf(jobParams.analysis['75-84%'].payableRate)
        if (jobParams.analysis['85-94%'] != null)
            return BigDecimal.valueOf(jobParams.analysis['85-94%'].payableRate)
        if (jobParams.analysis['95-99%'] != null)
            return BigDecimal.valueOf(jobParams.analysis['95-99%'].payableRate)
        if (jobParams.analysis['100%'] != null)
            return BigDecimal.valueOf(jobParams.analysis['100%'].payableRate)
        if (jobParams.analysis['Repetitions'] != null)
            return BigDecimal.valueOf(jobParams.analysis['Repetitions'].payableRate)
        if (jobParams.analysis['X translated'] != null)
            return BigDecimal.valueOf(jobParams.analysis['X translated'].payableRate)
        return null
    }

    def GetVendors(def vendorParams) {
        def providerService = utils.getService("com.radzisz.xtrf.service.ProviderService")
        def providers = providerService.getAll()

        List<Provider> filteredProviders = []
        if (vendorParams.filterActiveStatus)
            filteredProviders = providers.findAll{Provider provider -> provider.getStatus() == ProviderStatus.ACTIVE }
        else
            filteredProviders = providers.findAll{Provider provider -> provider}
            
         if (vendorParams.vendorIds)
            filteredProviders = filteredProviders.findAll{Provider provider -> vendorParams.vendorIds.contains(provider.getId().toString())}
            
        if (filteredProviders.size() > vendorParams.vendorLimit)
            filteredProviders = filteredProviders.take(vendorParams.vendorLimit + 1) // +1 is there to signal it was over the limit

        List<ProviderDTO> vendors = []
        filteredProviders.each{Provider provider -> vendors.add(getVendorData(provider)) }

        return vendors
    }

    def getVendorData(Provider provider) {
        return new ProviderDTO(
            provider.getId().toString(),
            provider.getName(),
            provider.getLastName(),
            provider.getFullname(),
            provider.getEmail(),
            provider.getStatus().toString(),
            getPriceProfileIds(provider)
        )
    }
    
    def getPriceProfileIds(Provider provider) {
        def priceProfiles = provider.getActivePriceProfiles()
        List<VendorPriceProfileDTO> priceProfileData = []
        priceProfiles.each{ProviderPriceProfile priceProfile -> priceProfileData.add( new VendorPriceProfileDTO(priceProfile.getId(), priceProfile.getName())) }
        return priceProfileData
    }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Testing params for the XTRF Editor. Uncomment if you want to test there, for production always keep it commented!
// You can specify the function you want to test, and the params needed for it.
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// def params = [function: "GetVendors", filterActiveStatus: true, vendorLimit: 2]

new BeLazyMacro(list, params, utils).runMacro()

class ProviderDTO {
    def externalId
    def userName
    def lastName
    def fullName
    def email
    def status
    def priceProfiles

    ProviderDTO(externalId, userName, lastName, fullName, email, status, priceProfiles) {
        this.externalId = externalId
        this.userName = userName
        this.lastName = lastName
        this.fullName = fullName
        this.email = email
        this.status = status
        this.priceProfiles = priceProfiles
    }
}

class VendorPriceProfileDTO {
    def id
    def name

    VendorPriceProfileDTO(id, name) {
        this.id = id
        this.name = name
    }
}